////////////////////////////////////////////////////////////////////////////////
// CONFIDENTIAL and PROPRIETARY software of Magewell Electronics Co., Ltd.
// Copyright (c) 2011-2019 Magewell Electronics Co., Ltd. (Nanjing)
// All rights reserved.
// This copyright notice MUST be reproduced on all authorized copies.
////////////////////////////////////////////////////////////////////////////////

#import "ViewController.h"
#import <mach/mach.h>
#import <mach/mach_time.h>

static uint64_t getTimestamp() {
    uint64_t time = mach_absolute_time();
    static mach_timebase_info_data_t sTimebaseInfo;
    
    if (sTimebaseInfo.denom == 0) {
        mach_timebase_info(&sTimebaseInfo);
    }
    
    long double ret = 1.L * time / 1000 * sTimebaseInfo.numer / sTimebaseInfo.denom;
    return ret / 1000;
}

static void onVideoCaptureCallback(BYTE *pBuffer, long bufferLen, long stride, void* pParam) {
    ViewController *viewCtrl = (__bridge ViewController*)pParam;
    [viewCtrl onVideoCaptureFrame:pBuffer bufferLen:bufferLen stride:stride];
}

static void onAudioCaptureCallback(const BYTE * pbFrame, int cbFrame, uint64_t u64TimeStamp, void* pParam) {
    ViewController *viewCtrl = (__bridge ViewController*)pParam;
    [viewCtrl onAudioCaptureFrame:(BYTE *)pbFrame cbFrame:cbFrame u64TimeStamp:u64TimeStamp];
}

@implementation ViewController
- (void)onVideoCaptureFrame:(BYTE *)pBuffer bufferLen:(long)BufferLen stride:(long)stride {
    //render
    if (self.viewEnable) {
        CVPixelBufferRef pixelBuffer = NULL;
        NSDictionary *pixelAttributes = @{(NSString*)kCVPixelBufferIOSurfacePropertiesKey:@{}};
        CVReturn result = CVPixelBufferCreate(kCFAllocatorDefault,
                                         self.width,
                                         self.height,
                                         self.pixelFormat,
                                         (__bridge CFDictionaryRef)(pixelAttributes),
                                         &pixelBuffer);

        CVPixelBufferLockBaseAddress(pixelBuffer,0);
        size_t pixelStride = CVPixelBufferGetBytesPerRowOfPlane(pixelBuffer, 0);
        unsigned char *yDestPlane = (unsigned char *)CVPixelBufferGetBaseAddressOfPlane(pixelBuffer, 0);
        unsigned char *uvDestPlane = (unsigned char *)CVPixelBufferGetBaseAddressOfPlane(pixelBuffer, 1);
        unsigned char *y_ch0 = pBuffer;
        unsigned char *y_ch1 = pBuffer + stride * self.height;
        if (pixelStride == stride) {
            memcpy(yDestPlane, y_ch0, stride * self.height);
            memcpy(uvDestPlane, y_ch1, stride * self.height / 2);
        } else {
            for (int i = 0; i < self.height; i++) {
                memcpy(yDestPlane + i*pixelStride, y_ch0 + i*stride, self.width);
            }
            for (int i = 0; i < self.height/2; i++) {
                memcpy(uvDestPlane + i*pixelStride, y_ch1+ i*stride, self.width);
            }
        }


        CVPixelBufferUnlockBaseAddress(pixelBuffer, 0);

        CMSampleTimingInfo timing = {kCMTimeInvalid, kCMTimeInvalid, kCMTimeInvalid};
        CMVideoFormatDescriptionRef videoInfo = NULL;
        result = CMVideoFormatDescriptionCreateForImageBuffer(NULL, pixelBuffer, &videoInfo);

        CMSampleBufferRef sampleBuffer = NULL;
        result = CMSampleBufferCreateForImageBuffer(kCFAllocatorDefault, pixelBuffer, true, NULL, NULL, videoInfo, &timing, &sampleBuffer);
        CFRelease(videoInfo);

        CFArrayRef attachments = CMSampleBufferGetSampleAttachmentsArray(sampleBuffer, YES);
        CFMutableDictionaryRef dict = (CFMutableDictionaryRef)CFArrayGetValueAtIndex(attachments, 0);
        CFDictionarySetValue(dict, kCMSampleAttachmentKey_DisplayImmediately, kCFBooleanTrue);

        if (self.videoLayer) {
            [self.videoLayer enqueueSampleBuffer:(CMSampleBufferRef)sampleBuffer];
        }

        CFRelease(sampleBuffer);
        CFRelease(pixelBuffer);
    }

    //fps
    {
        uint64_t now = getTimestamp();
        if (self.lastTimestamp == 0) {
            self.lastTimestamp = now;
            self.videoFrameCount = 0;
        } else {
            self.videoFrameCount++;
        }

        if (now - self.lastTimestamp >= 1000) {
            self.captureFps = self.videoFrameCount * 1000.0 / (now - self.lastTimestamp);
            self.lastTimestamp = now;
            self.videoFrameCount = 0;
            
            //NSLog(@"video capture fps: %f ", self.captureFps);
        }
    }
}

- (void)onAudioCaptureFrame:(BYTE *)pbFrame cbFrame:(int)cbFrame u64TimeStamp:(uint64_t)u64TimeStamp {
    MWAudioRenderPut(self.audioRender, (uint8_t*)pbFrame, cbFrame);

    //static FILE *fd = fopen("capture.pcm", "wb");
    //fwrite(pbFrame, cbFrame, 1, fd);
}

- (void)startCapture:(NSMenuItem *)deviceItem {
    [self stopCapture];

    int index = (int)[deviceItem tag];
    char wPath[256] = {0};
    MWGetDevicePath(index, wPath);
    HCHANNEL hChannel = MWOpenChannelByPath(wPath);
    if (hChannel == NULL) {
        printf("ERROR: Open channel failed !\n");
        return;
    }
    
    {
        bool resolutionFound = false;
        MWCAP_VIDEO_RESOLUTION resolutions[256] = { 0 };
        MWCAP_VIDEO_RESOLUTION_LIST resList = { sizeof(resolutions)/sizeof(resolutions[0]), resolutions};
        MWGetVideoCaptureSupportListResolution(hChannel, &resList);
        for(int i=0; i<resList.nListSize; i++) {
            printf("INFO: support resolution[%d x %d]\n", resolutions[i].cx, resolutions[i].cy);
            if (resolutions[i].cx == self.width &&
                resolutions[i].cy == self.height) {
                resolutionFound = true;
                //break;
            }
        }
        
        if (!resolutionFound) {
            printf("ERROR: resolution[%d x %d] not found !\n", self.width, self.height);
            MWCloseChannel(hChannel);
            return;
        }
    }
    
    //Audio Node menu
    [self initAudioMenu:hChannel];
    
    //Audio render
    {
        MWAudioRenderCreate(self.samplerate, self.audioChannels, self.bitsPerSample, &(_audioRender));
        MWAudioRenderStart(self.audioRender);
    }

    //AV Capture
    {
        self.audioCapture = MWCreateAudioCapture(hChannel, MWCAP_AUDIO_CAPTURE_NODE_DEFAULT, self.samplerate, self.bitsPerSample, self.audioChannels,
                                              onAudioCaptureCallback, (__bridge void *)(self));
        
        char szVideoPath[512] = { 0 };
        if(MW_SUCCEEDED == MWGetVideoDevicePathByChannel(hChannel, szVideoPath)) {
            //close channel first before video capture
            MWCloseChannel(hChannel);
            self.videoCapture = MWCreateVideoCapture(szVideoPath, self.width, self.height, self.fourcc, self.frameDuration,
                                                  onVideoCaptureCallback, (__bridge void *)(self));
        } else {
            MWCloseChannel(hChannel);
        }
    }
    
    [deviceItem setState:NSControlStateValueOn];
    self.selectedDeviceItem = deviceItem;
}

- (void)stopCapture {
    if (self.audioCapture) {
        MWDestoryAudioCapture(self.audioCapture);
        self.audioCapture = NULL;
    }
    
    if (self.audioRender) {
        MWAudioRenderStop(self.audioRender);
        MWAudioRenderDestroy(self.audioRender);
        self.audioRender = NULL;
    }
    
    if (self.videoCapture) {
        MWDestoryVideoCapture(self.videoCapture);
        self.videoCapture = NULL;
    }
    
    if (self.selectedDeviceItem) {
        [self.selectedDeviceItem setState:NSControlStateValueOff];
        self.selectedDeviceItem = NULL;
    }
    
    if (self.selectedAudioItem) {
        [self.selectedAudioItem setState:NSControlStateValueOff];
        self.selectedAudioItem = NULL;
    }
}

- (void)switchAudioNode:(NSMenuItem *)audioItem {
    if (self.audioCapture) {
        MWDestoryAudioCapture(self.audioCapture);
        self.audioCapture = NULL;
    }
    
    if (self.selectedAudioItem) {
        [self.selectedAudioItem setState:NSControlStateValueOff];
        self.selectedAudioItem = NULL;
    }
    
    int deviceIndex = (int)[self.selectedDeviceItem tag];
    char wPath[256] = {0};
    MWGetDevicePath(deviceIndex, wPath);
    HCHANNEL hChannel = MWOpenChannelByPath(wPath);
    if (hChannel == NULL) {
        printf("ERROR: Open channel failed !\n");
        return;
    }
    
    MWCAP_AUDIO_CAPTURE_NODE audioNode = (MWCAP_AUDIO_CAPTURE_NODE)[audioItem tag];
    self.audioCapture = MWCreateAudioCapture(hChannel, audioNode, self.samplerate, self.bitsPerSample, self.audioChannels, onAudioCaptureCallback, (__bridge void *)(self));
    
    MWCloseChannel(hChannel);
    hChannel = NULL;
    
    [audioItem setState:NSControlStateValueOn];
    self.selectedAudioItem = audioItem;
}

- (void)initParameters {
    self.width         = 1920;
    self.height        = 1080;
    self.frameDuration = 166667;
    self.fourcc        = MWFOURCC_NV12;

    self.samplerate    = 48000;
    self.audioChannels = 2;
    self.bitsPerSample = 16;

    self.audioRender   = NULL;
    self.audioCapture  = NULL;
    self.videoCapture  = NULL;

    self.captureFps      = 0.0;
    self.lastTimestamp   = 0;
    self.videoFrameCount = 0;
    
    self.pixelFormat = kCVPixelFormatType_420YpCbCr8BiPlanarVideoRange;
}

- (void)initDeviceMenu {
    int count = MWGetChannelCount();
    for (int i = 0; i < count; i++) {
        NSString *itemTitle = nil;
        
        MWCAP_CHANNEL_INFO mci = { 0 };
        MWGetChannelInfoByIndex(i, &mci);
        if (strcmp(mci.szFamilyName, "USB Capture") == 0) {
            itemTitle = [NSString stringWithFormat:@"%s(%s)", mci.szProductName, mci.szBoardSerialNo];
        } else {
            itemTitle = [NSString stringWithFormat:@"%02d-%d %s", mci.byBoardIndex, mci.byChannelIndex, mci.szProductName];
        }
        
        NSMenuItem *deviceItem = [self.deviceMenu addItemWithTitle:itemTitle action:@selector(startCapture:) keyEquivalent:@""];
        [deviceItem setTag:i];
    }
}

- (void)initAudioMenu:(HCHANNEL) hChannel {
    [self.audioMenu removeAllItems];

    MWCAP_CHANNEL_INFO mci = { 0 };
    MWGetChannelInfo(hChannel, &mci);
    if (0 == strcmp(mci.szFamilyName, "Pro Capture")) {
        NSString *itemTitle = [NSString stringWithFormat:@"%02d-%d %s", mci.byBoardIndex, mci.byChannelIndex, mci.szProductName];
        NSMenuItem *audioItem = [self.audioMenu addItemWithTitle:itemTitle action:@selector(switchAudioNode:) keyEquivalent:@""];
        [audioItem setTag:MWCAP_AUDIO_CAPTURE_NODE_DEFAULT];

        [audioItem setState:NSControlStateValueOn];
        self.selectedAudioItem = audioItem;
    } else if (strcmp(mci.szFamilyName, "USB Capture") == 0) {
        MWCAP_AUDIO_CAPS t_Caps;
        MWGetAudioCaps(hChannel, &t_Caps);
        if (t_Caps.dwCaps & MWCAP_USB_AUDIO_CAP_EMBEDDED_CAPTURE) {
            NSString *itemTitle = [NSString stringWithFormat:@"Digital (%s(%s))", mci.szProductName, mci.szBoardSerialNo];
            NSMenuItem *audioItem = [self.audioMenu addItemWithTitle:itemTitle action:@selector(switchAudioNode:) keyEquivalent:@""];
            [audioItem setTag:MWCAP_AUDIO_CAPTURE_NODE_EMBEDDED_CAPTURE];

            [audioItem setState:NSControlStateValueOn];
            self.selectedAudioItem = audioItem;
        }

        if (t_Caps.dwCaps & MWCAP_USB_AUDIO_CAP_MICROPHONE) {
            NSString *itemTitle = [NSString stringWithFormat:@"Mic (%s(%s))", mci.szProductName, mci.szBoardSerialNo];
            NSMenuItem *audioItem = [self.audioMenu addItemWithTitle:itemTitle action:@selector(switchAudioNode:) keyEquivalent:@""];
            [audioItem setTag:MWCAP_AUDIO_CAPTURE_NODE_MICROPHONE];
        }

        if (t_Caps.dwCaps & MWCAP_USB_AUDIO_CAP_USB_CAPTURE) {
            NSString *itemTitle = [NSString stringWithFormat:@"Computer (%s(%s))", mci.szProductName, mci.szBoardSerialNo];
            NSMenuItem *audioItem = [self.audioMenu addItemWithTitle:itemTitle action:@selector(switchAudioNode:) keyEquivalent:@""];
            [audioItem setTag:MWCAP_AUDIO_CAPTURE_NODE_USB_CAPTURE];
        }

        if (t_Caps.dwCaps & MWCAP_USB_AUDIO_CAP_LINE_IN)  {
            NSString *itemTitle = [NSString stringWithFormat:@"Analog (%s(%s))", mci.szProductName, mci.szBoardSerialNo];
            NSMenuItem *audioItem = [self.audioMenu addItemWithTitle:itemTitle action:@selector(switchAudioNode:) keyEquivalent:@""];
            [audioItem setTag:MWCAP_AUDIO_CAPTURE_NODE_LINE_IN];
        }
    }
}

- (void)createPreviewLayer {
    if (!self.videoLayer) {
        self.videoLayer = [[AVSampleBufferDisplayLayer alloc] init];
        self.videoLayer.frame = CGRectMake(0, self.statusLabel.frame.size.height,
                                           self.view.frame.size.width, self.view.frame.size.height - self.statusLabel.frame.size.height);
        self.videoLayer.bounds = CGRectMake(0, self.statusLabel.frame.size.height,
                                            self.view.bounds.size.width, self.view.bounds.size.height - self.statusLabel.frame.size.height);
        self.videoLayer.autoresizingMask = NSViewWidthSizable | NSViewHeightSizable;
        self.videoLayer.videoGravity = AVLayerVideoGravityResizeAspect;
        self.videoLayer.backgroundColor = [NSColor blackColor].CGColor;
        [[self.view layer] addSublayer:self.videoLayer];
        
        self.view.window.delegate = self;
    }
}

- (void)parseArgs {
    NSArray *args = [[NSProcessInfo processInfo] arguments];
    for (int i=0; i<args.count; i++) {
        NSString *item = [args objectAtIndex:i];
        const char *ptr = strstr([item UTF8String], "-w=");
        if (ptr) {
            self.width = atoi(ptr+strlen("-w="));
            continue;
        }
        
        ptr = strstr([item UTF8String], "-h=");
        if (ptr) {
            self.height = atoi(ptr+strlen("-h="));
            continue;
        }
        
        ptr = strstr([item UTF8String], "-f=");
        if (ptr) {
            self.frameDuration = atoi(ptr+strlen("-f="));
            continue;
        }
        
        ptr = strstr([item UTF8String], "-v=");
        if (ptr) {
           self.viewEnable = atoi(ptr+strlen("-v="));
           continue;
        }
    }
}

- (void)viewDidLoad {
    [self.view setWantsLayer:YES];
    [super viewDidLoad];
    // Do any additional setup after loading the view.
    
    //base UI
    {
        self.viewEnable = 1;
        [self performSelector:@selector(createPreviewLayer) withObject:nil afterDelay:0];
        
        NSMenu *mainMenu = [NSApp mainMenu];
        
        NSMenuItem *mainMenuDeviceItem = [[NSMenuItem alloc] init];
        [mainMenu addItem:mainMenuDeviceItem];
        self.deviceMenu = [[NSMenu alloc] initWithTitle:@"Device"];
        [mainMenuDeviceItem setSubmenu:self.deviceMenu];
        
        NSMenuItem *mainMenuAudioItem = [[NSMenuItem alloc] init];
        [mainMenu addItem:mainMenuAudioItem];
        self.audioMenu = [[NSMenu alloc] initWithTitle:@"Audio"];
        [mainMenuAudioItem setSubmenu:self.audioMenu];
    }
    
    MWCaptureInitInstance();
    MWRefreshDevice();
    
    [self initParameters]; [self parseArgs];
    [self initDeviceMenu];
    
    if (self.deviceMenu.numberOfItems > 0) {
        [self performSelector:@selector(startCapture:) withObject:[self.deviceMenu itemAtIndex:0] afterDelay:0];
    }
    
    self.statusTimer = [NSTimer scheduledTimerWithTimeInterval:0.5 target: self selector:@selector(statusTimerProc:) userInfo:nil repeats:YES];
}

- (BOOL)windowShouldClose:(NSWindow *)sender {
    [self stopCapture];
    
    if (self.statusTimer) {
        [self.statusTimer invalidate];
    }
    
    MWCaptureExitInstance();
    
    return YES;
}

- (void)viewDidDisappear {
}

- (void)setRepresentedObject:(id)representedObject {
    [super setRepresentedObject:representedObject];

    // Update the view, if already loaded.
}

- (void) statusTimerProc: (NSTimer *) timer {
    char strFourcc[5] = { 0 };
    *(uint32_t*)strFourcc = self.fourcc;
    [self.statusLabel setStringValue:[NSString stringWithFormat:@" %d x %d %s %.2f FPS ", self.width, self.height, strFourcc, self.captureFps]];
}

@end
